#!/usr/bin/perl -w
#
#getAUC.pl 1.3
# Obtains the AUC, ROC curve, and precision at several recall for various go categories,
# and outputs the median for all the measures.
#
# Created Thu Aug 10,2006 by Lourdes Pena Castillo (lourdes dot pena at gmail.com)
# $Id: getAUC.pl,v 1.17 2006/12/11 04:54:05 lourdes Exp lourdes $
# Copyright (C) 2006  Lourdes Pena Castillo
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# http://www.gnu.org/copyleft/gpl.html
#################################################################################
#################################################################################
use English;
use strict;

my $start = time();

my($goMatrix_file, $goCategories_file, $genes_file, $VERBOSE) = @ARGV; #Read arguments

#  Check values of arguments
if (!(defined($goMatrix_file) && defined($goCategories_file) && defined($genes_file))) {
    die ("Insufficient arguments. Usage: getAUC.pl goMatrix_file goCategories_file genes_file [verbose-level]\n");
}

if (!(defined($VERBOSE) && $VERBOSE =~ /^\d$/)) {
 $VERBOSE = 0; #0 - quiet mode, 1 - moderate, 2 - extreme
} 

##############################################################################
#  Global variables
my @GO_categories = (); #GO categories to be considered
my %goMatrix = ();      #GO Annotations per gene (hash key = gene ID) 
my %posGO = ();         #Column index of the GO categories of interest in go Matrix
my $sep = "\t";         #delimiter in goMatrix and for output
my @goAUC;              #AUC per GO category
my @means = ();              #means of measures
my @stdev = ();              #sd. of measures
my %sawGenes = ();      #Genes of interest

###########################################################
#Read in the GO categories to evaluate
#GO categories should be in a text file one per line and contain GO id "GO:\d{7}"

open(GOs, $goCategories_file) or die ("open($goCategories_file): $!\n");

while (my $line = <GOs>) {
    chomp($line);
    push (@GO_categories, $1) if ($line =~ /(GO:\d{7})/ && $1); # Extract GO id
}

close(GOs);

if (!@GO_categories) { die ("Error, found no GO categories in $goCategories_file \n") };

print "Found ", $#GO_categories + 1, " GO categories in $goCategories_file \n" if $VERBOSE;

###########################################################
#Read in the Genes IDs of interest
#Gene IDs should be in a text file one per line and contain Gene id

open(Genes, $genes_file) or die ("open($genes_file): $!\n");

while (my $line = <Genes>) {
    chomp($line);
    $line =~ s/^\s+//;		# Delete leading whitespace
    $line =~ s/\"//g; #remove quotes
    $sawGenes{$1} = 1 if ($line =~ /^(\S+)|\s,$|/ && $1); # Extract Gene id
}

close(Genes);

if (!%sawGenes) { die ("Error, found no Gene IDs in $genes_file \n") };

print "Found ", scalar keys %sawGenes, " genes in $genes_file \n" if $VERBOSE;

############################################################
#Read the goMatrix and kept only the columns for the GO categories in goCategories_file and the rows for the genes in genes_file
#First row contains the column names

open(GOMatrix, $goMatrix_file) or die ("open($goMatrix_file): $!\n");

{
 my $firstline = 1;
 my @nc = ();
 @posGO{@GO_categories} = ();

 while (my $line = <GOMatrix>) {
    chomp($line);
    my ($gene, @fields);
    if ($firstline) { #Assume first line contains column labels
       ($gene, @fields) = split($sep, $line);
       while (my $tmp = pop @fields) { #start from the last column
         if ($tmp =~ /(GO:\d{7})/ && exists $posGO{$1}){ #check whether GO category has to be evaluated
            unshift (@nc, $#fields+1); #Obtain column index in goMatrix
            $posGO{$1} = @GO_categories - @nc; #Obtain colum index in reduced goMatrix
         }
       }
       if (@nc < @GO_categories) {die ("Error, GO categories in $goCategories_file not in $goMatrix_file \n")};
       if ($VERBOSE > 1) {
         ($gene, @fields) = split($sep, $line);
         print join($sep, @nc), "\n";
         print join($sep, @fields[@nc]), "\n";
       }
       $firstline = 0;
    } else { #Process lines with annotations
      $line =~ s/\"//g; #remove quotes
      if ($line =~/^(\S+)$sep/) { #Get gene ID
        if (defined $sawGenes{$1}) { #Kept only those annotations for the genes in $genes_file
          ($gene, @fields) = split($sep, $line);
           $goMatrix{$gene} = [@fields[@nc]]; #insert in the hash table a reference to the array with the go annotations of interest
        }
      }
    }
 }
}

close(GOMatrix);

if (!%goMatrix) { die "Error, found no annotations in $goMatrix_file \n" };
print "Found ", scalar keys %goMatrix," genes in $goMatrix_file \n" if $VERBOSE;

###################################################
# For each GO category read the instance file and obtain ROC and AUC (following class reference formulation)
# For n go categories produce n different ROC graphs with area
# Instance file name should be GO#######_pred.txt

printf "%s\t%s\t%s\t%8s\t%8s\t%8s\t%8s\t%8s\t%8s\t%8s\t%8s\t%8s\n", ("Category", "P", "N", "P01R", "P10R", "P20R", "P50R", "P80R", "P100R", "AUC", "AUC_50", "Recall_0.01");

my $k = 1;

foreach my $go (keys %posGO) {
  print "Reading ${go}\n" if ($VERBOSE > 1);
  system("cp GO_empty.txt ${go}_pred.txt") if(! -e "${go}_pred.txt" and -e "GO_empty.txt"); #To add files filled with zeros
  (-e "${go}_pred.txt") or die("${go} is missing!\n");
  my @insts = &readInstances($go."_pred.txt", \%goMatrix, $posGO{$go}, \%sawGenes); #returns an arrays of references to arrays
  my $tmp = pop(@insts); #Extract last item with positive instances count
  my $P = $$tmp[1]; # Total Number of positives
  my $I = scalar @insts; # Total Number of instances
  my $N = $I -$P; #Total Number of Negatives
  if (!defined($P) or $P == 0) {
     print STDERR "Error, saw no positive instances of the class $go\n";
     next;
  } elsif (!defined($N) or $N == 0) {
     print STDERR "Error, saw no negative examples\n";
     next;
  }
  my @measures = &ROC(\@insts, $P, $N, $go); #Get AUC, precision at diff. recalls and a file with the ROC curve
  @measures = ($P, $N, @measures);
  push(@goAUC,[@measures]);
  
  # Calculate mean and variance using Knuth method
  # M1 = X1, Mk = Mk-1 + (Xk - Mk-1)/k ---average
  # S1 = 0, Sk = Sk-1 + (Xk - Mk-1) * (Xk-Mk) ---variance
  if ($k == 1) {
    @means = @measures;
    @stdev= (0,0,0,0,0,0,0,0,0,0,0);
  } else {
     for (my $i=0; $i <= $#measures; $i++) {
        my @oldMeans = @means;
        $means[$i] = $oldMeans[$i] + ($measures[$i] - $oldMeans[$i])/$k;
        $stdev[$i] = $stdev[$i] + ($measures[$i] - $oldMeans[$i]) * ($measures[$i] - $means[$i]);
     }
  }
  $k++;
  
  printf "%s\t", $go; &printArray(\@measures); 
}



{
 my $go_n = scalar @goAUC - 1;
 my $is_odd = $go_n % 2;

 if ( $go_n > 0) {
   my $n_measures = scalar @{$goAUC[0]};
   # Get std. deviation
   for (my $i =0; $i < $n_measures; $i++) {
     $stdev[$i] = sqrt($stdev[$i]/($go_n));
   }
   #Get the median of the measures
   my @medianValues =();
   for (my $i = 0; $i < $n_measures ; $i++) {
      @goAUC = sort { $b->[$i]  <=>  $a->[$i] } @goAUC;
      my $median;
      if (!$is_odd) {
        $median = $goAUC[($go_n / 2)]->[$i];
      } else {
        $median = $goAUC[int($go_n / 2)]->[$i] + (($goAUC[int($go_n / 2) + 1]->[$i] - $goAUC[int($go_n / 2)]->[$i]) / 2);
      }
      push(@medianValues, $median);
   }
   print "Median_values\t"; &printArray(\@medianValues);
   print "Mean_values\t"; &printArray(\@means);
   print "StDev_values\t"; &printArray(\@stdev);
 }
}

printf STDERR "Elapsed time: %d sec.\n", time() - $start;

exit;  
#end main program

#Subroutines

########### Print array in the right format
sub printArray{
 my @array = @{$_[0]};
 foreach my $num (@array) {
  printf "%.7f\t", $num;
 }
 print "\n";
}
####################################################
#####################################################
# Read in the instances
# Arguments: instance filename, goMatrix, column, eval genes
# Instance file contains Gene ID, confidence

sub readInstances {
 open(INSTS, $_[0]) or die "open($_[0]): $!\n";

 my @insts;
 my $inst_count = 0; #Count the positive instances 
 my $col_index = $_[2];
 my %eval_genes = %{$_[3]}; 

 while (my $line = <INSTS>) {
    chomp($line);
    $line =~ s/\"//g;	  # Delete quotes
    $line =~ s/^\s+//;		# Delete leading whitespace
    $line =~ s/\s+$//;		# Delete trailing whitespace
    my($instno, $prob) = split(/[\s,]+/, $line); #  Split fields on space character or comma
    if (not exists $eval_genes{$instno}) { next; } # Added to guarantee that only genes of interest are considered
    
    if ($_[1]{$instno}) { #Find out whether it belongs to the class according to the goMatrix
      my $goAnnotations = $_[1]{$instno}; #Get the annotations in the goMatrix for the gene instno
      ##Added for novel annotations to exclude previously annotated (PA) genes, In the novel matrix -2 = PA gene in GO term
      next if ($$goAnnotations[$col_index] == -2); 
      push(@insts, [$instno, $prob, $$goAnnotations[$col_index]] ); 
      $inst_count += $$goAnnotations[$col_index];
    } else { #if it is not in the goMatrix assume is negative
      push (@insts, [$instno, $prob, 0] );
    }
 }
 push(@insts, ["P", $inst_count]); #Total number of positive instances

 close(INSTS);

 return @insts;
}

############################################

### Inserted from ROC.pl (ROC_algs-1.0)
### Modified Aug 10, 2006 by Lourdes Pena Castillo <<<<<
#
#    ROC --- blah blah blah
#    Created Thu Jun  9 2005 by Tom Fawcett (<tfawcett@acm.org>)
#    $Copyright = "Copyright (2006) Tom Fawcett";
#
# ROC_algs is free software.  By using the ROC algorithms here, you agree to
# accept the following license agreement.
# ROC_algs (Beta Version) License Agreement
# All of the documentation and software included in the ROC_algs (Beta Version) 
# Software is copyrighted by Tom Fawcett.
# Copyright [2004-06] Tom Fawcett. All rights reserved. 
# Contact: Tom Fawcett, <tfawcett@acm.org>
#
# If you use this code in your research, I request that you reference the
# following article:
#
#  Tom Fawcett (2003). "ROC Graphs: Notes and Practical Considerations for
#  Researchers", HP Labs Tech Report HPL-2003-4.  Available from:
#  http://www.purl.org/net/tfawcett/papers/ROC101.pdf
#
# Redistribution and use, with or without modification, are permitted provided
# that the redistributions of code retain this license and the following
# disclaimer:
# -----
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL TOM FAWCETT
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# -----
#
# The Statistics::Descriptive and Statistics::Distributions modules are free
# software libraries.  You can redistribute them and/or modify them under the
# same terms as Perl itself.  See:
# http://www.cpan.org/modules/by-module/Statistics/
# for more details.


sub ROC {

 use constant DEBUG => 0;

 my ($insts, $P, $N, $pos_class) = @_; 
 my $PRODUCE_ROC	   = 1; #If 0 doesn't output the files with the ROC points
 my($TP, $FP)		    = (0,0);
 my(@ROC_points)	    = ([0,0,"inf"]);
 my($lastprob)		    = "inf";
 my($FPR, $TPR)		    = (0.0, 0.0);
 my($old_FPR, $old_TPR)	    = (0,0);
 my $AUC                    = 0;
 my $firstnFP = 50;
 my $AUC_top = 0; #AUC for the first nFP false positives
 my $specificity = 0;
 my $recall_01 = 0;
 my @measures = ();
 my @recallPoints = (0.01, 0.10, 0.20, 0.50, 0.80, 1);
 my $i = 0;
 my ($oldTP, $oldFP) = (0,0);

 #  Sort instances decreasing by probability
 @$insts = sort { $b->[1]  <=>  $a->[1] } @$insts;

 print join(", ", "Gene", "Confidence", "Annotation"), "\n" if ($VERBOSE > 1);
 
 for my $one_inst (@$insts) {
    print join(", ", $$one_inst[0], $$one_inst[1], $$one_inst[2]), "\n" if ($VERBOSE > 1);
    my($instno, $prob, $true_class) = @$one_inst;
    if ($true_class) {
	      $TP++;
    } else {
	      $FP++;
    }
    #  Update rates.  We know that $P != 0 and $N != 0.
    $TPR = $TP / $P; # the same as recall or sensitivity
    $FPR = $FP / $N;
    $specificity = ($N - $FP)/ $N;
    print "${FP}, specificity = ${specificity}, recall = ${TPR}, \n" if ($VERBOSE > 1);
    if (1 - $specificity <= 0.01 ) {
       $recall_01 = $TPR;
    }
    
    #Keep precision at the recall points - Added by LPC August 2006
    #The precision corresponds to the highest precision at the specific recall.
    while ($i < scalar @recallPoints && $old_TPR < $recallPoints[$i] && $TPR >= $recallPoints[$i]) {
         my $precision = 0.0;
         my $x = 0;
         if ($TPR == $recallPoints[$i]) { #if exact, take the value from TP and FP
            $precision = $TP/ ($TP + $FP);
         } elsif ($old_TPR  > 0) {  #do interpolation as in Davis & Goadrich, ICML 2006.
            $x = $recallPoints[$i] * $P - $oldTP;
            $precision = ($oldTP + $x) / ($oldTP + $x + $oldFP + ($oldFP - $FP)/($oldTP - $TP) * $x);
         } else { #make precision at previous recalls equal to the current precision
            $precision = $TP/ ($TP + $FP);
         }
         push (@measures, $precision);
         print "Recall at $recallPoints[$i], Precision = $precision, x = $x \n" if ($VERBOSE > 1);
         $i++;
    }
    
    $lastprob = $prob if ($lastprob eq "inf"); #We initialize lastprob to the first score we find 10/12/06
    #  This test ensures that we wait for the end of a span of
    #  equal probabilities before emitting a point.
    if ($prob != $lastprob) {
	        #  Add in trapezoid area
	        my($new_area) = &trapezoid_area($old_FPR, $old_TPR, $FPR, $TPR);
	       if (DEBUG) {
	        print "Adding in trap area ($old_FPR, $old_TPR, $FPR, $TPR) = $new_area\n";
	       }
        $AUC += $new_area;
        if ( ($oldFP == $firstnFP -1) and  ($FP == $firstnFP)) { #keep AUC to the left of the first n FPs
          $AUC_top = $AUC;
        }
	       #  Record this point
	       push(@ROC_points, [$FPR,$TPR,$lastprob]) if $PRODUCE_ROC;
	       $old_FPR = $FPR;
	       $old_TPR = $TPR;
    } elsif ( ($oldFP == $firstnFP -1) and  ($FP == $firstnFP)) { #keep AUC to the left of the first n FPs
       my($new_area) = &trapezoid_area($old_FPR, $old_TPR, $FPR, $TPR);
       $AUC_top = $AUC + $new_area;
    }
    
    $lastprob = $prob;
    ($oldTP, $oldFP) = ($TP, $FP);
 }

 #  Done processing instances.
 #  Connect ROC curve to (1,1) if we're not there yet
 if ($old_FPR != 1 or $old_TPR != 1) {
    my $new_area = &trapezoid_area($old_FPR, $old_TPR, $FPR, $TPR);
    print "Adding in trap area ($old_FPR, $old_TPR, $FPR, $TPR) = $new_area\n" if DEBUG;
    $AUC += $new_area;
    push(@ROC_points, [1,1,0])   if $PRODUCE_ROC;
 }

 if ($PRODUCE_ROC) {
    &output_ROC_points($pos_class, @ROC_points)
 }

 push(@measures, $AUC, $AUC_top, $recall_01);
 @measures;
}

sub output_ROC_points {
    my($P_class, @points) = @_;
    my($outfile) = "ROC.${P_class}";
    open(OUT, ">$outfile") or die "open(>$outfile): $!";
     # FPR TPR Prob
    foreach my $pt (@points) {
  	   printf OUT "%.7f\t%.7f\t%s\n", $pt->[0], $pt->[1], $pt->[2];
    }

    close(OUT);
}

#  This area assumes x1<=x2 and y1<=y2
#  This is simply base * 1/2 height.
sub trapezoid_area {
    my($x1, $y1, $x2, $y2) = @_;
    ( $x2 - $x1 ) * ( $y1 + ($y2 - $y1)/2.0 )
}

##### End of insertion from ROC.pl






